import cactus_schema.runner.uri as uri
from cactus_schema.runner.schema import (
    ClientInteraction,
    ClientInteractionType,
    CriteriaEntry,
    CSIPAusVersion,
    DataStreamPoint,
    DERCapabilityInfo,
    DERSettingsInfo,
    DERStatusInfo,
    EndDeviceMetadata,
    InitResponseBody,
    PreconditionCheckEntry,
    RequestData,
    RequestEntry,
    RequestList,
    RunGroup,
    RunnerStatus,
    RunRequest,
    StartResponseBody,
    StepEventStatus,
    StepStatus,
    TestCertificates,
    TestConfig,
    TestDefinition,
    TestUser,
    TimelineDataStreamEntry,
    TimelineStatus,
)

# The same ProceedResponse Model is used for responses from both the runner AND the orchestrator.
# It is defined once in `cactus_schema.orchestrator` and re-exported here for clients of the runner.
from cactus_schema.orchestrator import ProceedResponse

__all__ = [
    "uri",
    "ClientInteraction",
    "ClientInteractionType",
    "CriteriaEntry",
    "CSIPAusVersion",
    "DataStreamPoint",
    "DERCapabilityInfo",
    "DERSettingsInfo",
    "DERStatusInfo",
    "EndDeviceMetadata",
    "InitResponseBody",
    "PreconditionCheckEntry",
    "ProceedResponse",
    "RequestData",
    "RequestEntry",
    "RequestList",
    "RunnerStatus",
    "RunGroup",
    "RunRequest",
    "StartResponseBody",
    "StepEventStatus",
    "StepStatus",
    "TestCertificates",
    "TestConfig",
    "TestDefinition",
    "TestUser",
    "TimelineDataStreamEntry",
    "TimelineStatus",
]



from dataclasses import dataclass


@dataclass
class ComplianceClass:
    name: str
    description: str


# This is an adapted version of TS 5573 Table 12.5 - Applicability of tests to classes of DER client
COMPLIANCE_CLASS_ORDERED: list[tuple[str, ComplianceClass]] = [
    ("A", ComplianceClass("A", "All clients managing DER (Excluding demand response).")),
    ("G", ComplianceClass("G", "Clients managing generation-type or storage-type DER.")),
    ("L", ComplianceClass("L", "Clients managing load-type or storage-type DER.")),
    ("C", ComplianceClass("C", "Clients conforming with the optional ConnectionPoint extension.")),
    ("S-L", ComplianceClass("S-L", "Clients implementing Subscription/Notification functionality AND L class.")),
    ("S-G", ComplianceClass("S-G", "Clients implementing Subscription/Notification functionality AND G class.")),
    ("M", ComplianceClass("M", "Clients supporting management of sets of DER.")),
    ("DER-A", ComplianceClass("DER-A", "All DER.")),
    ("DER-G", ComplianceClass("DER-G", "All DER capable of generation.")),
    ("DER-L", ComplianceClass("DER-L", "All DER capable of consumption.")),
    ("DR-A", ComplianceClass("DR-A", "All clients managing demand response devices.")),
    ("DR-D", ComplianceClass("DR-D", "Clients managing or incorporated into DRED demand response devices.")),
    (
        "DR-L",
        ComplianceClass(
            "DR-L", "Clients managing load-type or storage-type products with demand response capabilities."
        ),
    ),
    (
        "DR-G",
        ComplianceClass(
            "DR-G", "Clients managing generation-type or storage-type products with demand response capabilities."
        ),
    ),
]


def fetch_compliance_classes(class_names: set[str]) -> list[ComplianceClass]:
    items: list[ComplianceClass] = []
    matched_keys: set[str] = set()

    for key, cc in COMPLIANCE_CLASS_ORDERED:
        if key in class_names:
            items.append(cc)
            matched_keys.add(key)

    # Now find anything leftover in class_names that we don't have a description for
    for class_name in class_names:
        if class_name not in matched_keys:
            items.append(ComplianceClass(class_name, ""))

    return items



import cactus_schema.orchestrator.uri as uri
from cactus_schema.orchestrator.schema import (
    HEADER_GROUP_ID,
    HEADER_GROUP_NAME,
    HEADER_RUN_ID,
    HEADER_TEST_ID,
    HEADER_USER_NAME,
    AdminStatsResponse,
    CSIPAusVersionResponse,
    GenerateClientCertificateRequest,
    InitRunRequest,
    InitRunResponse,
    PaginatedType,
    Pagination,
    PlaylistRunInfo,
    ProceedResponse,
    RunArtifactMultipleRequest,
    RunGroupRequest,
    RunGroupResponse,
    RunGroupUpdateRequest,
    RunResponse,
    RunStatusResponse,
    StartRunResponse,
    TestProcedureResponse,
    TestProcedureRunSummaryResponse,
    UserConfigurationRequest,
    UserConfigurationResponse,
    UserUpdateRequest,
    UserWithRunGroupsResponse,
)

__all__ = [
    "uri",
    "HEADER_GROUP_ID",
    "HEADER_GROUP_NAME",
    "HEADER_RUN_ID",
    "HEADER_TEST_ID",
    "HEADER_USER_NAME",
    "AdminStatsResponse",
    "CSIPAusVersionResponse",
    "GenerateClientCertificateRequest",
    "InitRunRequest",
    "InitRunResponse",
    "PaginatedType",
    "Pagination",
    "PlaylistRunInfo",
    "ProceedResponse",
    "RunArtifactMultipleRequest",
    "RunGroupRequest",
    "RunGroupResponse",
    "RunGroupUpdateRequest",
    "RunResponse",
    "RunStatusResponse",
    "StartRunResponse",
    "TestProcedureResponse",
    "TestProcedureRunSummaryResponse",
    "UserConfigurationRequest",
    "UserConfigurationResponse",
    "UserUpdateRequest",
    "UserWithRunGroupsResponse",
]



import cactus_schema.notification.uri as uri
from cactus_schema.notification.schema import (
    CollectedHeader,
    CollectedNotification,
    CollectEndpointResponse,
    ConfigureEndpointRequest,
    CreateEndpointResponse,
)

__all__ = [
    "uri",
    "CollectedHeader",
    "CollectedNotification",
    "CollectEndpointResponse",
    "ConfigureEndpointRequest",
    "CreateEndpointResponse",
]



from dataclasses import dataclass
from datetime import datetime
from enum import StrEnum, auto
from typing import Generic, TypeVar

from cactus_test_definitions.client import TestProcedureId
from dataclass_wizard import JSONWizard
from dataclass_wizard.enums import LetterCase

HEADER_USER_NAME = "CACTUS-User-Name"
HEADER_TEST_ID = "CACTUS-Test-Id"
HEADER_RUN_ID = "CACTUS-Run-Id"
HEADER_GROUP_ID = "CACTUS-Group-Id"
HEADER_GROUP_NAME = "CACTUS-Group-Name"


class RunStatusResponse(StrEnum):
    initialised = auto()
    started = auto()
    finalised = auto()
    provisioning = auto()
    skipped = auto()


PaginatedType = TypeVar("PaginatedType")


class FastAPICompatibleWizard(JSONWizard):
    """This is our way of generating JSON that FastAPI should happily interact with directly as a dataclass"""

    class Meta(JSONWizard.Meta):
        key_transform_with_dump = LetterCase.SNAKE
        key_transform_with_load = LetterCase.SNAKE
        encode_enum_as_value = True
        datetime_to = "iso"


@dataclass
class Pagination(Generic[PaginatedType]):
    total_pages: int
    total_items: int
    page_size: int
    current_page: int
    prev_page: int | None
    next_page: int | None

    items: list[PaginatedType]


@dataclass
class InitRunRequest(FastAPICompatibleWizard):
    test_procedure_id: TestProcedureId | None = (
        None  # Single test (for backwards compatibility only, use test_procedure_ids)
    )
    test_procedure_ids: list[TestProcedureId] | None = None  # List of tests (playlist if >1)
    start_index: int | None = None  # 0-based index to start execution from (for playlists)


@dataclass
class PlaylistRunInfo(FastAPICompatibleWizard):
    """Summary info for a run within a playlist"""

    run_id: int
    test_procedure_id: str
    status: RunStatusResponse


@dataclass
class StartRunResponse(FastAPICompatibleWizard):
    test_url: str


@dataclass
class InitRunResponse(StartRunResponse):
    run_id: int
    playlist_execution_id: str | None = None  # Set if this is a playlist
    playlist_runs: list[PlaylistRunInfo] | None = None  # All runs in playlist order


@dataclass
class RunResponse(FastAPICompatibleWizard):
    run_id: int
    test_procedure_id: str
    test_url: str
    status: RunStatusResponse
    all_criteria_met: bool | None  # Whether this run has been assessed as successful or not (None means unknown)
    created_at: datetime
    finalised_at: datetime | None
    is_device_cert: bool  # Whether this run was initialised with the device cert or aggregator cert
    has_artifacts: bool  # Whether run has artifacts available (False if not finalized or failure during finalization)
    playlist_execution_id: str | None = None  # Set if this run is part of a playlist
    playlist_order: int | None = None  # 0-based position in playlist
    playlist_runs: list[PlaylistRunInfo] | None = None  # All runs in playlist with status


@dataclass
class RunArtifactMultipleRequest(FastAPICompatibleWizard):
    run_ids: list[int]


@dataclass
class RunGroupRequest(FastAPICompatibleWizard):
    csip_aus_version: str


@dataclass
class GenerateClientCertificateRequest(FastAPICompatibleWizard):
    is_device_cert: bool


@dataclass
class RunGroupUpdateRequest(FastAPICompatibleWizard):
    """NOTE - this is explicitly NOT allowing updates on csip-aus version - it has too many weird considerations and
    realistically, a user should just create a new group if they want to test against a new version (there is no
    practical need to allow migrating legacy version test runs to a newer version)"""

    name: str | None  # If non null - update the RunGroup receiving this request


@dataclass
class RunGroupResponse(FastAPICompatibleWizard):
    run_group_id: int
    name: str
    csip_aus_version: str
    created_at: datetime

    is_device_cert: bool | None
    certificate_id: int | None
    certificate_created_at: datetime | None

    total_runs: int  # How many runs live underneath this group


@dataclass
class UserWithRunGroupsResponse(FastAPICompatibleWizard):
    """Represents a user with all their associated run groups"""

    user_id: int
    subject_id: str
    name: str | None
    run_groups: list[RunGroupResponse]


@dataclass
class CSIPAusVersionResponse(FastAPICompatibleWizard):
    """Represents the various CSIP-Aus versions available for testing"""

    version: str  # Derived from the cactus_test_definitions.CSIPAusVersion enum


@dataclass
class TestProcedureResponse(FastAPICompatibleWizard):
    __test__ = False
    test_procedure_id: TestProcedureId
    description: str
    category: str
    classes: list[str]


@dataclass
class TestProcedureRunSummaryResponse(FastAPICompatibleWizard):
    __test__ = False
    test_procedure_id: TestProcedureId
    description: str
    category: str
    classes: list[str] | None
    run_count: int  # Count of runs for this test procedure
    latest_all_criteria_met: bool | None  # Value for all_criteria_met of the most recent Run
    latest_run_status: int | None  # RunStatus of the most recent Run
    latest_run_id: int | None  # run_id of the most recent Run
    latest_run_timestamp: datetime | None  # timestamp of the most recent Run


@dataclass
class AdminStatsResponse(FastAPICompatibleWizard):
    total_runs: int
    max_run_id: int
    total_passed: int
    total_failed: int
    total_users: int
    total_run_groups: int
    version_counts: dict[str, int]  # csip_aus_version -> count
    runs_per_week: dict[str, int]  # "2026-W07" -> count
    runs_per_user: dict[str, int]  # user_name -> count
    procedures: list[dict]  # top 20, each a plain dict


@dataclass
class UserConfigurationRequest(FastAPICompatibleWizard):
    subscription_domain: str | None  # What domain will outgoing notifications be scoped to? If None - no update
    is_static_uri: (
        bool | None
    )  # If true - all test instances will share the same URI (limit to 1 test at a time). If None - no update
    pen: int | None


@dataclass
class UserUpdateRequest(FastAPICompatibleWizard):
    user_name: str


@dataclass
class UserConfigurationResponse(FastAPICompatibleWizard):
    subscription_domain: str  # What domain will outgoing notifications be scoped to? Empty string = no value configured
    is_static_uri: bool  # If true - all test instances will share the same URI (limit to 1 test at a time).
    pen: int
    static_uri: str | None  # What the static URI will be for this user (readonly and only set if is_static_uri is True)


@dataclass
class ProceedResponse(FastAPICompatibleWizard):
    handled: bool  # If true, the proceed signal matched a listener and moved the test to the next step



from dataclasses import dataclass, field
from datetime import datetime
from enum import Enum, StrEnum, auto
from http import HTTPMethod, HTTPStatus

from cactus_test_definitions import CSIPAusVersion
from cactus_test_definitions.client import TestProcedureId
from dataclass_wizard import JSONWizard


class ClientInteractionType(StrEnum):
    RUNNER_START = "Runner Started"
    TEST_PROCEDURE_INIT = "Test Procedure Initialised"
    TEST_PROCEDURE_START = "Test Procedure Started"
    PROXIED_REQUEST = "Request Proxied"
    TEST_PROCEDURE_FINALIZED = "TEST_PROCEDURE_FINALIZED"


class StepStatus(Enum):
    PENDING = 0  # The step is not yet active
    ACTIVE = auto()  # The step is currently active but not complete
    RESOLVED = auto()  # The step has been full resolved


@dataclass
class RequestEntry(JSONWizard):
    url: str
    path: str
    method: HTTPMethod
    status: HTTPStatus
    timestamp: datetime
    step_name: str
    body_xml_errors: list[str]  # Any XML schema errors detected in the incoming body
    request_id: int  # Increments per test


@dataclass
class InitResponseBody(JSONWizard):
    status: str
    test_procedure: str
    timestamp: datetime
    is_started: bool = (
        False  # True if the run has progressed to the started state. False if it's still waiting for a call to start it
    )


@dataclass
class CriteriaEntry(JSONWizard):
    success: bool
    type: str
    details: str


@dataclass
class PreconditionCheckEntry(JSONWizard):
    success: bool
    type: str
    details: str


@dataclass
class DataStreamPoint(JSONWizard):
    watts: int | None  # The data point value (in watts)
    offset: str  # Label for identifying the relative start - usually something like "2m20s"


@dataclass
class TimelineDataStreamEntry(JSONWizard):
    label: str  # Descriptive label of this data stream
    data: list[DataStreamPoint]
    stepped: bool  # If True - this data should be presented as a stepped line chart
    dashed: bool  # If True - this data should be a dashed line


@dataclass
class TimelineStatus(JSONWizard):
    data_streams: list[TimelineDataStreamEntry]  # The set of data streams that should be rendered on the timeline
    set_max_w: int | None  # The currently set set_max_w (if any)
    now_offset: str  # The name of the DataStreamPoint.offset that corresponds with "now" (when this was calculated)


@dataclass
class DERCapabilityInfo(JSONWizard):
    """Snapshot of DERCapability for UI display"""

    der_type: str | None = None  # e.g. "PHOTOVOLTAIC_SYSTEM", "COMBINED_PV_AND_STORAGE"
    modes_supported: list[str] | None = None  # Active DERControlType flag names
    max_w: int | None = None
    max_va: int | None = None
    max_var: int | None = None
    max_var_neg: int | None = None
    max_a: int | None = None
    max_charge_rate_w: int | None = None
    max_discharge_rate_w: int | None = None
    max_wh: int | None = None
    doe_modes_supported: list[str] | None = None  # Active DOESupportedMode flag names


@dataclass
class DERSettingsInfo(JSONWizard):
    """Snapshot of DERSettings for UI display"""

    modes_enabled: list[str] | None = None  # Active DERControlType flag names
    max_w: int | None = None
    max_va: int | None = None
    max_var: int | None = None
    max_var_neg: int | None = None
    max_charge_rate_w: int | None = None
    max_discharge_rate_w: int | None = None
    grad_w: int | None = None  # Ramp rate (%setMaxW/second)
    doe_modes_enabled: list[str] | None = None  # Active DOESupportedMode flag names


@dataclass
class DERStatusInfo(JSONWizard):
    """Snapshot of current DER real-time status (from SiteDERStatus / sep2 DERStatus).
    Bitmaps/enums resolved to strings."""

    alarm_status: list[str] | None = None  # Active AlarmStatusType flag names
    inverter_status: str | None = None  # InverterStatusType enum name
    operational_mode_status: str | None = None  # OperationalModeStatusType enum name
    generator_connect_status: list[str] | None = None  # Active ConnectStatusType flag names
    storage_connect_status: list[str] | None = None  # Active ConnectStatusType flag names
    storage_mode_status: str | None = None  # StorageModeStatusType enum name
    state_of_charge_status: int | None = None  # Percent
    local_control_mode_status: str | None = None  # LocalControlModeStatusType enum name
    manufacturer_status: str | None = None  # Up to 6 chars


@dataclass
class EndDeviceMetadata(JSONWizard):  # All optional as a device may not always be registered
    edevid: int | None = None  # Should always be 1, but nice to check
    lfdi: str | None = None
    sfdi: int | None = None
    nmi: str | None = None
    aggregator_id: int | None = None
    set_max_w: int | None = None
    doe_modes_enabled: int | None = None
    device_category: int | None = None
    timezone_id: str | None = None
    der_capability: DERCapabilityInfo | None = None
    der_settings: DERSettingsInfo | None = None
    der_status: DERStatusInfo | None = None


@dataclass
class StartResponseBody(JSONWizard):
    status: str
    test_procedure: str
    timestamp: datetime


@dataclass
class RequestData(JSONWizard):
    request_id: int
    request: str | None
    response: str | None


@dataclass
class ClientInteraction(JSONWizard):
    interaction_type: ClientInteractionType
    timestamp: datetime


@dataclass
class StepEventStatus:
    started_at: datetime | None  # When was this step event handler enabled
    completed_at: datetime | None  # When was this step event handler completed at
    event_status: str | None = None  # Status update from the event listener for this step (eg - "Waiting 30 seconds")


@dataclass
class RunnerStatus(JSONWizard):
    timestamp_status: datetime  # when was this status generated?
    timestamp_initialise: datetime | None  # When did the test initialise
    timestamp_start: datetime | None  # When did the test start
    status_summary: str
    last_client_interaction: ClientInteraction
    csip_aus_version: str  # The CSIPAus version that is registered in the active test procedure (can be empty)
    log_envoy: str  # Snapshot of the current envoy logs
    criteria: list[CriteriaEntry] = field(default_factory=list)
    precondition_checks: list[PreconditionCheckEntry] = field(default_factory=list)
    instructions: list[str] | None = field(default=None)
    test_procedure_name: str = field(default="-")  # '-' represents no active procedure
    step_status: dict[str, StepEventStatus] | None = field(default=None)
    request_history: list[RequestEntry] = field(default_factory=list)
    timeline: TimelineStatus | None = None  # Streaming timeline data snapshot
    end_device_metadata: EndDeviceMetadata | None = None  # Snapshot of current active end device (if any)


@dataclass
class RequestList(JSONWizard):
    request_ids: list[int]
    count: int


@dataclass
class TestDefinition(JSONWizard):
    test_procedure_id: TestProcedureId
    yaml_definition: str


@dataclass
class TestCertificates(JSONWizard):
    aggregator: str | None
    device: str | None


@dataclass
class RunGroup(JSONWizard):
    run_group_id: str
    name: str
    csip_aus_version: CSIPAusVersion
    test_certificates: TestCertificates


@dataclass
class TestConfig(JSONWizard):
    subscription_domain: str | None
    is_static_url: bool
    pen: int = field(default=0)


@dataclass
class TestUser(JSONWizard):
    user_id: str
    name: str


@dataclass
class RunRequest(JSONWizard):
    run_id: str
    test_definition: TestDefinition
    run_group: RunGroup
    test_config: TestConfig
    test_user: TestUser



#
# GET - Returns a text/plain response with server status
# POST Not Supported
# PUT - Not Supported
# DELETE - Not Supported
#
URI_MANAGE_SERVER = "/manage"

#
# GET - Not Supported
# POST CreateEndpointRequest -> CreateEndpointResponse
# PUT - Not Supported
# DELETE - Not Supported
#
URI_MANAGE_ENDPOINT_LIST = "/manage/endpoint"

#
# GET -> CollectEndpointResponse
# POST Not Supported
# PUT ConfigureEndpointRequest -> No Body
# DELETE -> No body
#
URI_MANAGE_ENDPOINT = "/manage/endpoint/{endpoint_id}"


#
# Listens and logs any method
#
URI_ENDPOINT = "/webhook/{endpoint_id}"



from dataclasses import dataclass
from datetime import datetime

from dataclass_wizard import JSONWizard


@dataclass(frozen=True)
class CreateEndpointResponse(JSONWizard):
    """Used in response to a CreateEndpointRequest"""

    endpoint_id: str  # Unique ID for identifying and managing the endpoint
    fully_qualified_endpoint: str  # A fully qualified https:// endpoint that can be used for subscription webhook


@dataclass(frozen=True)
class CollectedHeader(JSONWizard):
    """Simple representation of a HTTP header"""

    name: str
    value: str


@dataclass(frozen=True)
class CollectedNotification(JSONWizard):
    """Simple representation of a HTTP notification submitted by the utility server to a notification endpoint"""

    method: str  # What HTTP method was used
    headers: list[CollectedHeader]  # The submitted HTTP headers
    body: str  # String encoded HTTP request body
    received_at: datetime  # tz aware datetime
    remote: str | None  # Who sent this request (typically an IP address)


@dataclass(frozen=True)
class CollectEndpointResponse(JSONWizard):
    """Used to respond to a CollectEndpointRequest"""

    notifications: list[CollectedNotification]


@dataclass(frozen=True)
class ConfigureEndpointRequest(JSONWizard):
    """Used for adjusting the configuration of an existing endpoint"""

    enabled: (
        bool  # if False - submitted requests to the endpoint will be served with a HTTP 500 and ignored. Defaults True
    )



AdminUsersList = "/admin/users"
AdminRunGroupProceduresList = "/admin/procedure_runs/{run_group_id}"
AdminRunGroupList = "/admin/run_group"
AdminRunArtifact = "/admin/run/{run_id}/artifact"
AdminRunGroupProcedureRunList = "/admin/procedure_runs/{run_group_id}/{test_procedure_id}"
AdminRunGroupRunList = "/admin/run_group/{run_group_id}/run"
AdminRunRegenerateReport = "/admin/run/{run_id}/regenerate"
AdminRunStatus = "/admin/run/{run_id}/status"
AdminRun = "/admin/run/{run_id}"
AdminRunGroupCompliance = "/admin/run_group/{run_group_id}/artifact"
AdminStats = "/admin/stats"

CertificateAuthority = "/certificate/authority"
CertificateRunGroup = "/run_group/{run_group_id}/certificate"
CertificateRunGroups = "/run_group/certificate"

Config = "/config"

VersionList = "/version"
ProcedureList = "/procedure"
Procedure = "/procedure/{test_procedure_id}"
ProcedureRunGroupList = "/procedure_runs/{run_group_id}"
ProcedureRunGroupRunsList = "/procedure_runs/{run_group_id}/{test_procedure_id}"

RunGroupList = "/run_group"
RunGroup = "/run_group/{run_group_id}"

RunGroupRunList = "/run_group/{run_group_id}/run"
Run = "/run/{run_id}"
RunFinalise = "/run/{run_id}/finalise"
RunPlaylistFinalise = "/run/{run_id}/finalise_playlist"
RunArtifact = "/run/{run_id}/artifact"
RunStatus = "/run/{run_id}/status"
RunRequestList = "/run/{run_id}/requests"
RunRequest = "/run/{run_id}/requests/{request_id}"
RunArtifactMultiple = "/run/artifact/multiple"
RunProceed = "/run/{run_id}/proceed"

User = "/user"
